#!/usr/bin/env bash

# This script has the lofty goal of becoming a full configuration utility for mutt.
# Written by Storm Dragon: https://social.stormdragon.tk/storm
# Written by Michael Taboada: https://2mb.social/mwtab
# Contributions by Kyle: https://kyle.tk
# Released under the terms of  the WTFPL: http://wtfpl.net

# Settings to improve accessibility of dialog.
export DIALOGOPTS='--insecure --no-lines --visit-items'

# Use the cli entry system for gpg keys.
export PINENTRY_USER_DATA="USE_CURSES=1"

# Array of command line arguments
declare -A command=(
    [h]="Show this help information."
    [t]="Test mode. Use ~/mutt_test instead of ~/.mutt for configuration."
)

# Variables
muttHome=~/.mutt
fileDir="$(dirname "$(realpath "$0")")/files"
testMode=false

# Process command line arguments
# Convert the keys of the associative array to a format usable by getopts
args="${!command[*]}"
args="${args//[[:space:]]/}"
while getopts "${args}" i ; do
    case "$i" in
        h) help ;;
        t) 
           testMode=true
           muttHome=~/mutt_test
           ;;
    esac
done

# Functions
help() {
    echo "fleacollar.sh"
    echo "Released under the terms of the WTFPL: http://wtfpl.net"
    echo -e "A configuration utility for mutt.\n"
    echo -e "Usage:\n"
    echo "With no arguments, run the configuration utility."
    for i in "${!command[@]}" ; do
        echo "-${i}: ${command[${i}]}"
    done | sort
    exit 0
}

check_dependancies()
{
    local dep
    for dep in dialog gpg2 mutt pass w3m ; do
       if ! command -v $dep &> /dev/null ; then
            echo "$dep is not installed. Please install $dep and run this script again."
            exit 1
         fi
    done
    if ! [ -d ~/.gnupg ]; then
        read -p "No configuration for GPG was found. To have Fleacollar configure this for you press enter. If you would like to configure GPG manually, press control+c. " continue
        configure_gpg
    fi
}

inputbox() {
    # Returns: text entered by the user
    # Args 1, Instructions for box.
    # args: 2 initial text (optional)
    dialog --backtitle "Enter text and press enter." \
        --inputbox "$1" 0 0 "$2" --stdout
}
 
menulist() {
    # Args: minimum group 2, multiples of 2, "tag" "choice"
    # returns: selected tag
    local menuList
    ifs="$IFS"
    IFS=$'\n'
    dialog --backtitle "Use the up and down arrow keys to find the option you want, then press enter to select it." \
        --no-tags \
        --menu "Please select one" 0 0 0 $@ --stdout
    IFS="$ifs"
}

msgbox() {
# Returns: None
# Shows the provided message on the screen with an ok button.
dialog --msgbox "$*" 0 0
} 

passwordbox() {
    # Returns: text entered by the user
    # Args 1, Instructions for box.
    # args: 2 initial text (optional)
    dialog --backtitle "Enter text and press enter." \
        --passwordbox "$1" 0 0 "$2" --stdout
}
 
yesno() {
    # Returns: Yes or No
    # Args: Question to user.
    # Called  in if $(yesno) == "Yes"
    # Or variable=$(yesno)
    dialog --backtitle "Press 'Enter' for \"yes\" or 'Escape' for
\"no\"." --yesno "$*" 0 0 --stdout
    if [[ $? -eq 0 ]]; then
        echo "Yes"
    else
        echo "No"
    fi
}
 
initialize_directory()
{
    mkdir -p "${muttHome}/cache/bodies"
    mkdir -p "${muttHome}/scripts"
    if ! [ -f "$muttHome/aliases" ]; then
        touch "$muttHome/aliases"
    fi
    
    # Copy the add_address.sh script
    if ! [ -f "$muttHome/scripts/add_address.sh" ]; then
        cp "$fileDir/add_address.sh" "$muttHome/scripts/add_address.sh"
        chmod 700 "$muttHome/scripts/add_address.sh"
    fi
    
    # Create mailcap file (now always using w3m)
    if ! [ -f "$muttHome/mailcap" ]; then
        cp "$fileDir/mailcap" "$muttHome/mailcap"
    fi
    
    # Create gpg.rc file
    if ! [ -f "$muttHome/gpg.rc" ]; then
        find /usr -name gpg.rc -exec cp "{}" "$muttHome/" \; 2> /dev/null
        echo "set pgp_autosign=yes" >> "$muttHome/gpg.rc"
        echo "set crypt_autosign=yes" >> "$muttHome/gpg.rc"
        echo "set pgp_replyencrypt=yes" >> "$muttHome/gpg.rc"
        echo "set pgp_timeout=1800" >> "$muttHome/gpg.rc"
        if [[ $(gpg --list-secret-keys | wc -l) -eq 0 ]]; then
            read -p "No gpg key was found. Type your name and press enter to generate a PGP key. Control+c if you would like to create it manually. " continue
            # Try to use quick key generation, and fall back to the more verbose version for legacy systems.
            if ! gpg2 --quick-gen-key "${continue:-${USER}}"; then
                gpg2 --gen-key
            fi
        fi
        PS3="Select the key you want to use for encryption/signing:"
        select key in $(gpg --list-secret-keys | grep '.*@.*' | cut -d '<' -f2 | cut -d '>' -f1) ; do
            if [[ -n "$key" ]]; then
                break
            fi
        done
        keyName="$(gpg2 --list-secret-keys --keyid-format short | grep -B1 ^uid | head -n1 | rev | cut -c -8 | rev)"
        echo "set pgp_sign_as=$keyName" >> "$muttHome/gpg.rc"
    fi
    
    # Create or update macro file
    cp "$fileDir/macros" "$muttHome/macros"
    
    # Create basic muttrc
    if ! [ -f "$muttHome/muttrc" ]; then
        # Find desired editor
        x=0
        for i in nano vim ; do
            unset editorPath
            editorPath="$(command -v $i)"
            if [ -n "$editorPath" ]; then
                editors[$x]="$editorPath"
                ((x++))
            fi
        done
        echo "Select editor for email composition:"
        i="$(menulist ${EDITOR} ${EDITOR} $(for i in "${editors[@]##*/}" ; do echo "$i";echo "$i";done))"
        if [ -z "$i" ]; then
            exit 0
        fi
        
        # Copy base muttrc
        cp "$fileDir/muttrc" "$muttHome/muttrc"
        
        # Set editor in muttrc
        case "$i" in
            "nano") sed -i "1s|^|set editor = '$i +7 -r 72'\n|" "$muttHome/muttrc" ;;
            "vim") sed -i "1s|^|set editor = \"vim -c 'set spell spelllang=${LANG::2}'\"\n|" "$muttHome/muttrc" ;;
        esac
        
        # Replace muttHome in muttrc
        sed -i "s|\$muttHome|${muttHome/#$HOME/\~}|g" "$muttHome/muttrc"
    fi
}

configure_gpg()
{
    # GPG stuff in .bashrc:
    if ! grep 'GPG_TTY=$(tty)'  ~/.bashrc ; then
        echo -e 'GPG_TTY=$(tty)\nexport GPG_TTY' >> ~/.bashrc
    fi
    
    # Make sure the configuration directory exists
    if ! [ -d ~/.gnupg/ ]; then
        gpg2 --list-secret-keys &> /dev/null
    fi
    
    if [ -f ~/.gnupg/gpg.conf ]; then
        read -p "This will overwrite your existing ~/.gnupg/gpg.conf file. Press enter to continue or control+c to abort. " continue
    fi
    
    # Copy GPG configuration files
    cp "$fileDir/gpg.conf" ~/.gnupg/gpg.conf
    
    if [ -f ~/.gnupg/gpg-agent.conf ]; then
        read -p "This will overwrite your existing ~/.gnupg/gpg-agent.conf file. Press enter to continue or control+c to abort. " continue
    fi
    
    cp "$fileDir/gpg-agent.conf" ~/.gnupg/gpg-agent.conf
}

add_email_address()
{
    read -p "Please enter your email address: " emailAddress
    if ! [[ "$emailAddress" =~ .*@.*\..* ]]; then
        read -p "This appears to be an invalid email address. Continue anyway? (y/n) " continue
        if [ "${continue^}" != "Y" ]; then
            exit 0
        fi
    fi
    if [ -f "$muttHome/$emailAddress" ]; then
        read -p "This email address already exists. Overwrite the existing settings? (y/n) " continue
        if [ "${continue^}" != "Y" ]; then
            exit 0
        else
            sed -i "/$emailAddress/d" "$muttHome/muttrc"
        fi
    fi
    read -p "Enter your name as you want it to appear in emails. From: " realName
    echo "set realname=\"$realName\"" > "$muttHome/$emailAddress"
    echo "set from=\"$emailAddress\"" >> "$muttHome/$emailAddress"
    echo "set use_from = \"yes\"" >> "$muttHome/$emailAddress"
    echo "set hostname=${emailAddress##*@}" >> "$muttHome/$emailAddress"
    case "$emailAddress" in
        *gmail.com)
            configure_gmail "$emailAddress"
        ;;
        *hotmail.com)
            configure_hotmail "$emailAddress"
        ;;
        *)
            continue="$(yesno "Is this a gmail account?")"
            if [[ "$continue" == "Yes" ]]; then
                configure_gmail "$emailAddress"
            else
                configure_generic "$emailAddress"
            fi
    esac
    
    # Password storage with pass
    passOne=a
    passTwo=b
    until [ "$passOne" = "$passTwo" ]; do
        passOne="$(passwordbox "Please enter the password for $emailAddress:")"
        passTwo="$(passwordbox "Please enter the password again: ")"
        if [ "$passOne" != "$passTwo" ]; then
            echo "The passwords do not match."
        fi
    done
    
    # Store password in pass
    pass insert -f "email/$emailAddress" <<< "$passOne"
    
    # Add configuration to retrieve password from pass
    echo "set imap_pass=\"`pass show email/$emailAddress`\"" >> "$muttHome/$emailAddress"
    echo "set smtp_pass=\"`pass show email/$emailAddress`\"" >> "$muttHome/$emailAddress"
    
    add_keybinding
    echo "folder-hook .*$emailAddress/ 'source ${muttHome/#$HOME/\~}/$emailAddress'" >> "$muttHome/$emailAddress"
    msgbox "Email address added, press enter to continue."
}

configure_gmail()
{
    # Copy the Gmail template and replace placeholders
    cp "$fileDir/gmail.template" "$muttHome/$1.tmp"
    
    # Replace email placeholder in the template
    sed -i "s|EMAIL_ADDRESS|$1|g" "$muttHome/$1.tmp"
    sed -i "s|USERNAME|${1%@*}|g" "$muttHome/$1.tmp"
    
    # Append the template to the existing config
    cat "$muttHome/$1.tmp" >> "$muttHome/$1"
    rm "$muttHome/$1.tmp"
    
    # Handle goobook integration if available
    unset continue
    if command -v goobook &> /dev/null ; then
        read -p "Goobook is installed, would you like to use it as your addressbook for the account $1? " continue
        if [ "${continue^}" = "Y" ]; then
            echo "set query_command=\"goobook query %s\"" >> "$muttHome/$1"
            echo "macro index,pager a \"<pipe-message>goobook add<return>\" \"add sender to google contacts\"" >> "$muttHome/$1"
        else
            echo "source ~/${muttHome#/home/*/}/aliases" >> "$muttHome/$1"
        fi
    else
        echo "source ~/${muttHome#/home/*/}/aliases" >> "$muttHome/$1"
    fi
}

configure_hotmail()
{
    # Copy the Hotmail template and replace placeholders
    cp "$fileDir/hotmail.template" "$muttHome/$1.tmp"
    
    # Replace email placeholder in the template
    sed -i "s|EMAIL_ADDRESS|$1|g" "$muttHome/$1.tmp"
    
    # Append the template to the existing config
    cat "$muttHome/$1.tmp" >> "$muttHome/$1"
    rm "$muttHome/$1.tmp"
    
    # Add source for aliases
    echo "source ~/${muttHome#/home/*/}/aliases" >> "$muttHome/$1"
}

configure_generic()
{
    # Break the email address into its components:
    local userName="${1%%@*}"
    local hostName="${1##*@}"
    local imapHost
    local imapUser
    local imapPort
    local smtpHost
    local smtpUser
    local smtpPort
    
    # Gather input
    read -p "Enter imap host: " -e -i imap.$hostName imapHost
    read -p "Enter imap user: " -e -i $1 imapUser
    read -p "Enter imap port: " -e -i 993 imapPort
    read -p "Enter smtp host: " -e -i smtp.$hostName smtpHost
    read -p "Enter smtp user: " -e -i $userName smtpUser
    read -p "Enter smtp port: " -e -i 587 smtpPort
    
    # Copy the generic template
    cp "$fileDir/generic.template" "$muttHome/$1.tmp"
    
    # Replace placeholders
    sed -i "s|IMAP_HOST|$imapHost|g" "$muttHome/$1.tmp"
    sed -i "s|IMAP_USER|$imapUser|g" "$muttHome/$1.tmp"
    sed -i "s|IMAP_PORT|$imapPort|g" "$muttHome/$1.tmp"
    sed -i "s|SMTP_HOST|$smtpHost|g" "$muttHome/$1.tmp"
    sed -i "s|SMTP_USER|$smtpUser|g" "$muttHome/$1.tmp"
    sed -i "s|SMTP_PORT|$smtpPort|g" "$muttHome/$1.tmp"
    
    # Append the template to the existing config
    cat "$muttHome/$1.tmp" >> "$muttHome/$1"
    rm "$muttHome/$1.tmp"
    
    # Add any extra settings
    read -p "Enter extra settings, one line at a time, just press enter when done: " extraSettings
    while [ "$extraSettings" != "" ]; do
        echo "$extraSettings" >> "$muttHome/$1"
        read extraSettings
    done
    
    # Add source for aliases
    echo "source ~/${muttHome#/home/*/}/aliases" >> "$muttHome/$1"
}

add_keybinding()
{
    # Here we search for previous keybinding
    local fNumber=1
    while : ; do
        grep "^[[:space:]m]acro.*index.*<F$fNumber>.*" $muttHome/muttrc &> /dev/null || break # fNumber is now the currently open keybinding.
        ((fNumber++)) # fNumber was taken, so increment it.
    done
    # Bind key FfNumber to the mail account.
    echo "macro generic,index <F$fNumber> '<sync-mailbox><enter-command>source ${muttHome/#$HOME/\~}/$emailAddress<enter><change-folder>!<enter>'" >> "$muttHome/muttrc"
    echo "mail account $emailAddress bound to F$fNumber."
    if ! grep "^source.*@.*\..*" "$muttHome/muttrc" &> /dev/null ; then
        continue="$(yesno "Make $emailAddress the default account?")"
        if [ "$continue" = "Yes" ]; then
            echo "source ${muttHome/#$HOME/\~}/$emailAddress" >> "$muttHome/muttrc"
        fi
    fi
}

new_contact()
{
    contactName="$(inputbox "Enter the contact's first and last name..")"
    if [ -z "$contactName" ]; then
        exit 0
    fi
    contactEmail="$(inputbox "Enter the email address for $contactName")"
    if [ -z "$contactEmail" ]; then
        exit 0
    fi
    contactAlias="${contactName,,}"
    contactAlias="${contactAlias// /-}"
    if grep "<$contactEmail>\| $contactAlias " "$muttHome/aliases" &> /dev/null ; then
        [[ "$(yesno "This email address already exists in your contacts. Continue anyway?")" != "Yes" ]] && exit 0
    fi
    echo "alias $contactAlias $contactName <$contactEmail>" >> "$muttHome/aliases"
    sort -u "$muttHome/aliases" -o "$muttHome/aliases"
    msgbox "$contactName added to your address book."
}

# This is the main loop of the program
# Call functions to be ran every time the script is ran.
check_dependancies
initialize_directory

# If in test mode, display a notification
if [ "$testMode" = true ]; then
    echo "Running in test mode. Using $muttHome for configuration."
fi

# Let's make a mainmenu variable to hold all the options for the select loop.
mainmenu=("Add Email Address" "Configure GPG" "New Contact" "Exit")
while : ; do
    i="$(IFS=$'\n';menulist $(for i in "${mainmenu[@]}" ; do echo "$i";echo "$i";done))"
    [[ -z "$i" ]] && continue
    functionName="${i,,}"
    functionName="${functionName// /_}"
    functionName="${functionName/exit/exit 0}"
    $functionName
done

exit 0
